# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
import logging
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.units import Quantity
from gammapy.data import EventListMetaData, EventList, ObservationTable
from gammapy.utils.scripts import make_path
from gammapy.utils.metadata import CreatorMetaData
from gammapy.utils.time import time_ref_from_dict

log = logging.getLogger(__name__)


class EventListReader:
    """Reader class for EventList.

    Format specification: :ref:`gadf:iact-events`

    Parameters
    ----------
    hdu : str
        Name of events HDU. Default is "EVENTS".
    checksum : bool
        If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.
    """

    def __init__(self, hdu="EVENTS", checksum=False):
        self.hdu = hdu
        self.checksum = checksum

    @staticmethod
    def from_gadf_hdu(events_hdu):
        """Create EventList from gadf HDU."""
        table = Table.read(events_hdu)
        meta = EventListMetaData.from_header(table.meta)

        # This is not a strict check on input. It just checks that required information is there.
        required_colnames = set(["RA", "DEC", "TIME", "ENERGY"])
        if not required_colnames.issubset(set(table.colnames)):
            missing_columns = required_colnames.difference(set(table.colnames))
            raise ValueError(
                f"GADF event table does not contain required columns {missing_columns}"
            )

        met = u.Quantity(table["TIME"].astype("float64"), "second")
        time = time_ref_from_dict(table.meta) + met

        energy = table["ENERGY"].quantity

        ra = table["RA"].quantity
        dec = table["DEC"].quantity

        removed_colnames = ["RA", "DEC", "GLON", "GLAT", "TIME", "ENERGY"]

        new_table = Table(
            {"TIME": time, "ENERGY": energy, "RA": ra, "DEC": dec}, meta=table.meta
        )
        for name in table.colnames:
            if name not in removed_colnames:
                new_table.add_column(table[name])

        return EventList(new_table, meta)

    @staticmethod
    def identify_format_from_hduclass(events_hdu):
        """Identify format for HDU header keywords."""
        hduclass = events_hdu.header.get("HDUCLASS", "unknown")
        return hduclass.lower()

    def read(self, filename, format="gadf"):
        """Read EventList from file.

        Parameters
        ----------
        filename : `pathlib.Path`, str
            Filename
        format : {"gadf"}, optional
            format of the EventList. Default is 'gadf'.
            If None, will try to guess from header.
        """
        filename = make_path(filename)

        with fits.open(filename) as hdulist:
            events_hdu = hdulist[self.hdu]

            if self.checksum:
                if events_hdu.verify_checksum() != 1:
                    warnings.warn(
                        f"Checksum verification failed for HDU {self.hdu} of {filename}.",
                        UserWarning,
                    )

            if format is None:
                format = self.identify_format_from_hduclass(events_hdu)

            if format == "gadf" or format == "ogip":
                return self.from_gadf_hdu(events_hdu)
            else:
                raise ValueError(f"Unknown format :{format}")


class EventListWriter:
    """Writer class for EventList."""

    def __init__(self):
        pass

    @staticmethod
    def _to_gadf_table_hdu(event_list):
        """Convert input event list to a `~astropy.io.fits.BinTableHDU` according gadf."""
        gadf_table = event_list.table.copy()
        gadf_table.remove_column("TIME")

        reference_time = time_ref_from_dict(gadf_table.meta)
        gadf_table["TIME"] = (event_list.time - reference_time).to("s")

        bin_table = fits.BinTableHDU(gadf_table, name="EVENTS")

        # A priori don't change creator information
        if event_list.meta.creation is None:
            event_list.meta.creation = CreatorMetaData()
        else:
            event_list.meta.creation.update_time()

        bin_table.header.update(event_list.meta.to_header())
        return bin_table

    def to_hdu(self, event_list, format="gadf"):
        """
        Convert input event list to a `~astropy.io.fits.BinTableHDU` according to format.

        Parameters
        ----------
        format : str, optional
            Output format, currently only "gadf" is supported. Default is "gadf".

        Returns
        -------
        hdu : `astropy.io.fits.BinTableHDU`
            EventList converted to FITS representation.
        """
        if format != "gadf":
            raise ValueError(f"Only the 'gadf' format supported, got {format}")

        return self._to_gadf_table_hdu(event_list)


class ObservationTableReader:
    """Reader class for ObservationTable."""

    valid_formats = ["GADF", "OGIP"]

    def read(self, filename, hdu=None):
        """Read an Observation index table in GADF format and convert to internal `~gammapy.data.ObservationTable`.

        Parameters
        ----------
        filename : `pathlib.Path`, str
            Filename
        hdu : str, optional
            Name of observation table HDU. Default is None.

        Returns
        -------
        observation_table : `~gammapy.data.ObservationTable`
            Observation table.
        """
        filename = make_path(filename)

        table = Table.read(filename, format="fits", hdu=hdu)

        format = table.meta.get("HDUCLASS")

        if format is None:
            format = "GADF"
            log.warning(
                f"Could not infer the file format from metadata in {filename}, assuming GADF."
            )

        if format.upper() in ["GADF", "OGIP"]:
            return self._from_gadf_table(table)
        else:
            raise ValueError(
                f"Unknown file format :{format}. Expected one of {self.valid_formats}."
            )

    @staticmethod
    def _from_gadf_table(table_gadf):
        """Convert GADF observation table into `~gammapy.data.ObservationTable`.

        Parameters
        ----------
        table_gadf : `~astropy.Table.table`
            Table in GADF format (version 0.2 or 0.3).
            See [1]_, [2]_ for further information.

        Returns
        -------
        observation_table : `~gammapy.data.ObservationTable`
            Observation table.

        References
        ----------
        .. [1] Gamma-ray astronomy community. (2018), "Data formats for gamma-ray astronomy v0.2" '<https://gamma-astro-data-formats.readthedocs.io/en/v0.2/data_storage/obs_index/index.html>'_
        .. [2] Gamma-ray astronomy community. (2018), "Data formats for gamma-ray astronomy v0.3" '<https://gamma-astro-data-formats.readthedocs.io/en/v0.3/data_storage/obs_index/index.html>'_
        """
        names_gadf = table_gadf.colnames
        meta_gadf = table_gadf.meta

        required_names = [
            "OBS_ID",
        ]

        missing_names = set(required_names).difference(names_gadf)
        if missing_names:
            raise RuntimeError(
                f"The input table is missing required GADF columns. The missing columns are: {missing_names}."
            )

        time_columns = set(["TSTART", "TSTOP"]).intersection(set(names_gadf))

        if time_columns:
            try:
                time_ref = time_ref_from_dict(meta_gadf)
                time_unit = meta_gadf["TIMEUNIT"]
            except KeyError:
                log.warning(
                    "Found column TSTART or TSTOP in GADF table, but can not create columns in internal format due to missing header keywords in file."
                )
                table_gadf.remove_columns(time_columns)
                return ObservationTable(data=table_gadf, meta=meta_gadf)

            for colname in time_columns:
                try:
                    col_type_converted = table_gadf[colname].astype("float64")
                except ValueError:
                    log.warning(
                        f"Could not convert type for column {colname} to float64."
                    )
                    table_gadf.remove_column(colname)
                    break
                try:
                    q = Quantity(col_type_converted, time_unit)
                except ValueError:
                    log.warning(
                        f"Unit for column {colname} not understood. Unit with dimension of time expected."
                    )
                    table_gadf.remove_column(colname)
                    break
                try:
                    table_gadf[colname] = time_ref + q
                except TypeError:
                    log.warning(
                        f"Unit for column {colname} is in wrong dimension. Unit with dimension of time expected."
                    )
                    table_gadf.remove_column(colname)

        return ObservationTable(data=table_gadf, meta=meta_gadf)



"""Plot an energy dispersion from the HESS DL3 data release 1."""

import matplotlib.pyplot as plt
from gammapy.irf import EnergyDispersion2D

filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
edisp = EnergyDispersion2D.read(filename, hdu="EDISP")
edisp.peek()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import pytest
import numpy as np
import astropy.units as u
from numpy.testing import assert_allclose, assert_equal
from gammapy.datasets import Datasets, SpectrumDatasetOnOff
from gammapy.datasets.tests.test_map import get_map_dataset
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    FoVBackgroundModel,
    Models,
    SkyModel,
    PiecewiseNormSpectralModel,
)
from gammapy.modeling.tests.test_fit import MyDataset
from gammapy.stats import GaussianPriorPenalty
from gammapy.utils.testing import requires_data


@pytest.fixture(scope="session")
def datasets():
    return Datasets([MyDataset(name="test-1"), MyDataset(name="test-2")])


@pytest.fixture(scope="session")
def map_datasets():
    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=2)
    geom = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.05,
        width=(20, 20),
        frame="icrs",
        axes=[axis],
    )

    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=3, name="energy_true")
    geom_etrue = WcsGeom.create(
        skydir=(266.40498829, -28.93617776),
        binsz=0.05,
        width=(20, 20),
        frame="icrs",
        axes=[axis],
    )

    dataset_1 = get_map_dataset(geom, geom_etrue, name="test-1")
    dataset_1.mask_fit = None
    dataset_1.background /= 400
    dataset_2 = get_map_dataset(geom, geom_etrue, name="test-2")
    datasets = Datasets([dataset_1, dataset_2])

    model = SkyModel.create("pl", "point", name="src")
    model.spatial_model.position = dataset_1.exposure.geom.center_skydir

    model2 = model.copy()
    model2.spatial_model.lon_0.value += 0.1
    model2.spatial_model.lat_0.value += 0.1

    models = Models(
        [
            model,
            model2,
            FoVBackgroundModel(dataset_name=dataset_1.name),
            FoVBackgroundModel(dataset_name=dataset_2.name),
        ]
    )

    datasets.models = models
    dataset_1.fake(random_state=42)
    dataset_2.fake(random_state=42)
    map_datasets = Datasets([dataset_1, dataset_2])
    return map_datasets


@pytest.fixture(scope="session")
def heterogenous_map_datasets():
    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=2)
    geom = WcsGeom.create(skydir=(266.0, -28.0), binsz=0.05, width=(2, 2), axes=[axis])
    geom2 = WcsGeom.create(skydir=(266.0, -28.1), binsz=0.05, width=(2, 2), axes=[axis])

    axis = MapAxis.from_energy_bounds("0.1 TeV", "10 TeV", nbin=3, name="energy_true")
    geom_etrue = geom.to_image().to_cube([axis])

    dataset_1 = get_map_dataset(geom, geom_etrue, name="test-1")
    dataset_2 = get_map_dataset(geom2, geom_etrue, name="test-2")
    dataset_3 = dataset_1.copy()
    dataset_4 = dataset_2.copy()
    dataset_3.counts = None
    dataset_4.counts = None
    return Datasets([dataset_1, dataset_2, dataset_3, dataset_4])


def test_datasets_init(datasets):
    # Passing a Python list of `Dataset` objects should work
    Datasets(list(datasets))

    # Passing an existing `Datasets` object should work
    Datasets(datasets)


def test_datasets_types(datasets):
    assert datasets.is_all_same_type


def test_datasets_likelihood(datasets):
    likelihood = datasets.stat_sum()
    assert_allclose(likelihood, 14472200.0002)


@requires_data()
def test_datasets_likelihood_with_penalty(map_datasets):
    assert_allclose(map_datasets.stat_sum(), 4132.493313)

    norm_model = PiecewiseNormSpectralModel(energy=np.geomspace(0.1, 10, 5) * u.TeV)

    penalty = GaussianPriorPenalty.from_method(
        norm_model.parameters, "L2", mean=0.0, lambda_=2
    )

    map_datasets.models["src"].spectral_model *= norm_model

    models = Models(map_datasets.models, penalties=penalty)
    assert len(models._penalties) == 1
    models.set_penalties(penalty)
    assert len(models._penalties) == 1

    assert_allclose(penalty.stat_sum(), 10)

    map_datasets.models = models

    assert_allclose(map_datasets.stat_sum(), 4132.493313 + 10)

    map_datasets.models["src"].spectral_model = map_datasets.models[
        "src"
    ].spectral_model.model1
    models = Models(map_datasets.models, penalties=None)
    assert models._penalties == map_datasets.models._penalties
    models.set_penalties(None)
    assert models._penalties is None
    map_datasets.models = (
        models  # TODO: removing the penalty is quite complex should be simpler
    )

    assert_allclose(map_datasets.stat_sum(), 4132.493313)

    with pytest.raises(ValueError):
        models.set_penalties([[1, 2]])


def test_datasets_str(datasets):
    assert "Datasets" in str(datasets)


def test_datasets_getitem(datasets):
    assert datasets["test-1"].name == "test-1"
    assert datasets["test-2"].name == "test-2"


def test_names(datasets):
    assert datasets.names == ["test-1", "test-2"]


def test_datasets_mutation():
    dat = MyDataset(name="test-1")
    dats = Datasets([MyDataset(name="test-2"), MyDataset(name="test-3")])
    dats2 = Datasets([MyDataset(name="test-4"), MyDataset(name="test-5")])

    dats.insert(0, dat)
    assert dats.names == ["test-1", "test-2", "test-3"]

    dats.extend(dats2)
    assert dats.names == ["test-1", "test-2", "test-3", "test-4", "test-5"]

    dat3 = dats[3]
    dats.remove(dats[3])
    assert dats.names == ["test-1", "test-2", "test-3", "test-5"]
    dats.append(dat3)
    assert dats.names == ["test-1", "test-2", "test-3", "test-5", "test-4"]
    dats.pop(3)
    assert dats.names == ["test-1", "test-2", "test-3", "test-4"]

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.append(dat)

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.insert(0, dat)

    with pytest.raises(ValueError, match="Dataset names must be unique"):
        dats.extend(dats2)


@requires_data()
def test_datasets_info_table():
    datasets_hess = Datasets()

    for obs_id in [23523, 23526]:
        dataset = SpectrumDatasetOnOff.read(
            f"$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs{obs_id}.fits"
        )
        datasets_hess.append(dataset)

    table = datasets_hess.info_table()
    assert table["name"][0] == "23523"
    assert table["name"][1] == "23526"
    assert np.isnan(table["npred_signal"][0])
    assert_equal(table["counts"], [124, 126])
    assert_allclose(table["background"], [7.666, 8.583], rtol=1e-3)

    table_cumul = datasets_hess.info_table(cumulative=True)
    assert table_cumul["name"][0] == "stacked"
    assert table_cumul["name"][1] == "stacked"
    assert np.isnan(table_cumul["npred_signal"][0])
    assert table_cumul["alpha"][1] == table_cumul["alpha"][0]
    assert_equal(table_cumul["counts"], [124, 250])
    assert_allclose(table_cumul["background"], [7.666, 16.25], rtol=1e-3)

    assert table["excess"].sum() == table_cumul["excess"][1]
    assert table["counts"].sum() == table_cumul["counts"][1]
    assert table["background"].sum() == table_cumul["background"][1]

    datasets_hess[0].mask_safe.data = ~datasets_hess[0].mask_safe.data
    assert datasets_hess.info_table()["counts"][0] == 65

    datasets_hess[0].mask_safe.data = np.ones_like(
        datasets_hess[0].mask_safe.data, dtype=bool
    )
    datasets_hess[1].mask_safe.data = np.ones_like(
        datasets_hess[1].mask_safe.data, dtype=bool
    )
    assert_equal(datasets_hess.info_table()["counts"], [189, 199])
    assert_equal(datasets_hess.info_table(cumulative=True)["counts"], [189, 388])


@requires_data()
def test_datasets_write(map_datasets, tmp_path):
    datasets = Datasets(map_datasets[0])

    datasets.write(
        filename=tmp_path / "test",
        filename_models=tmp_path / "test_model",
        overwrite=False,
    )
    os.remove(tmp_path / "test-1.fits")

    with pytest.raises(OSError):
        datasets.write(
            filename=tmp_path / "test",
            filename_models=tmp_path / "test_model",
            overwrite=False,
        )

    with pytest.raises(ValueError, match="The filename is not defined."):
        datasets.write(filename=None)


@requires_data()
def test_datasets_fit(map_datasets):
    datasets = map_datasets

    fit = Fit()
    results = fit.run(datasets)

    assert_allclose(results.models.covariance.data, datasets.models.covariance.data)


def test_add_datasets(datasets):
    assert isinstance(datasets + Datasets(), Datasets)
    assert isinstance(Datasets() + datasets[0], Datasets)
    assert isinstance(datasets[0] + Datasets(), Datasets)
    assert isinstance(datasets[0] + datasets[1], Datasets)
    assert isinstance(datasets[0] + [], Datasets)

    assert len(datasets[0] + datasets[1]) == 2

    new_datasets = datasets[0] + datasets[1]
    assert len(new_datasets) == len(datasets.models)
    assert len(new_datasets) == 2

    tmp = np.ones(3)
    with pytest.raises(TypeError):
        datasets[0] + tmp

    with pytest.raises(TypeError):
        datasets + tmp


@requires_data()
def test_is_all_same_geom(map_datasets):
    assert map_datasets.is_all_same_geom


@requires_data()
def test_is_all_same_geom_false(heterogenous_map_datasets):
    assert not heterogenous_map_datasets.is_all_same_geom
    # Different counts geometries
    assert not Datasets(heterogenous_map_datasets[:2]).is_all_same_geom
    # No counts maps but different background map geometries
    assert not Datasets(heterogenous_map_datasets[2:]).is_all_same_geom



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from gammapy.scripts.main import cli
from gammapy.utils.testing import run_cli


def test_cli_info_help():
    result = run_cli(cli, ["info", "--help"])
    assert "Usage" in result.output


def test_cli_info_no_args():
    # No arguments should print all infos
    result = run_cli(cli, ["info"])
    assert "System" in result.output
    assert "Gammapy package" in result.output
    assert "Other packages" in result.output
    assert "Gammapy environment variables" in result.output
    assert "pydantic" in result.output
    assert "jupyter" in result.output
    assert "PyYAML" in result.output



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from itertools import repeat
import numpy as np
import scipy.stats as stats
from astropy import units as u
from astropy.table import Table
import gammapy.utils.parallel as parallel
from gammapy.datasets import Datasets
from gammapy.datasets.utils import set_and_restore_mask_fit
from gammapy.datasets.actors import DatasetsActor
from gammapy.datasets.flux_points import _get_reference_model
from gammapy.maps import MapAxis, Map
from gammapy.modeling import Fit, Sampler, Parameters, Parameter
from gammapy.modeling.models import (
    Models,
    PowerLawNormSpectralModel,
    TemplateNPredModel,
    UniformPrior,
)

from ..flux import FluxEstimator
from .core import FluxPoints

log = logging.getLogger(__name__)

__all__ = ["FluxPointsEstimator", "FluxCollectionEstimator"]


class FluxPointsEstimator(FluxEstimator, parallel.ParallelMixin):
    """Flux points estimator.

    Estimates flux points for a given list of datasets, energies and spectral model.

    To estimate the flux point the amplitude of the reference spectral model is
    fitted within the energy range defined by the energy group. This is done for
    each group independently. The amplitude is re-normalized using the "norm" parameter,
    which specifies the deviation of the flux from the reference model in this
    energy group. See https://gamma-astro-data-formats.readthedocs.io/en/latest/spectra/binned_likelihoods/index.html
    for details.

    The method is also described in the `Fermi-LAT catalog paper <https://ui.adsabs.harvard.edu/abs/2015ApJS..218...23A>`__
    or the `H.E.S.S. Galactic Plane Survey paper <https://ui.adsabs.harvard.edu/abs/2018A%26A...612A...1H>`__

    Parameters
    ----------
    source : str or int
        For which source in the model to compute the flux points.
    n_sigma : float, optional
        Number of sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Number of sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    n_sigma_sensitivity : float, optional
        Sigma to use for sensitivity computation. Must be a positive value.
        Default is 5.
    selection_optional : list of str, optional
        Which additional quantities to estimate. Available options are:

            * "all": all the optional steps are executed.
            * "errn-errp": estimate asymmetric errors on flux.
            * "ul": estimate upper limits.
            * "scan": estimate fit statistic profiles.
            * "sensitivity": estimate sensitivity for a given significance.

        Default is None so the optional steps are not executed.
    energy_edges : list of `~astropy.units.Quantity`, optional
        Edges of the flux points energy bins. The resulting bin edges won't be exactly equal to the input ones,
        but rather the closest values to the energy axis edges of the parent dataset.
        Default is [1, 10] TeV.
    fit : `~gammapy.modeling.Fit`, optional
        Fit instance specifying the backend and fit options. If None, the `~gammapy.modeling.Fit` instance is created
        internally. Default is None.
    reoptimize : bool, optional
        If True, the free parameters of the other models are fitted in each bin independently,
        together with the norm of the source of interest
        (but the other parameters of the source of interest are kept frozen).
        If False, only the norm of the source of interest is fitted,
        and all other parameters are frozen at their current values.
        Default is False.
    sum_over_energy_groups : bool, optional
        Whether to sum over the energy groups or fit the norm on the full energy grid. Default is None.
    n_jobs : int, optional
        Number of processes used in parallel for the computation. The number of jobs is limited to the number of
        physical CPUs. If None, defaults to `~gammapy.utils.parallel.N_JOBS_DEFAULT`.
        Default is None.
    parallel_backend : {"multiprocessing", "ray"}, optional
        Which backend to use for multiprocessing. If None, defaults to `~gammapy.utils.parallel.BACKEND_DEFAULT`.
    norm : `~gammapy.modeling.Parameter` or dict, optional
        Norm parameter used for the fit.
        Default is None and a new parameter is created automatically, with value=1, name="norm",
        scan_min=0.2, scan_max=5, and scan_n_values = 11. By default, the min and max are not set
        (consider setting them if errors or upper limits computation fails). If a dict is given,
        the entries should be a subset of `~gammapy.modeling.Parameter` arguments.
    allow_multiple_telescopes : bool, optional
        Whether to allow the computation for different telescopes.
        **WARNING**: This is currently an experimental feature.
        Default is False.

    Notes
    -----
    - For further explanation, see :ref:`estimators`.
    - In case of failure of upper limits computation (e.g. nan), see the User Guide: :ref:`dropdown-UL`.

    Examples
    --------
    .. testcode::

        from astropy import units as u
        from gammapy.datasets import SpectrumDatasetOnOff
        from gammapy.estimators import FluxPointsEstimator
        from gammapy.modeling.models import PowerLawSpectralModel, SkyModel

        path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
        dataset = SpectrumDatasetOnOff.read(path + "pha_obs23523.fits")

        pwl = PowerLawSpectralModel(index=2.7, amplitude='3e-11  cm-2 s-1 TeV-1')

        dataset.models = SkyModel(spectral_model=pwl, name="crab")

        estimator = FluxPointsEstimator(
            source="crab",
            energy_edges=[0.1, 0.3, 1, 3, 10, 30, 100] * u.TeV,
        )

        fp = estimator.run(dataset)
        print(fp)

    .. testoutput::

        FluxPoints
        ----------

          geom                   : RegionGeom
          axes                   : ['lon', 'lat', 'energy']
          shape                  : (1, 1, 6)
          quantities             : ['norm', 'norm_err', 'ts', 'npred', 'npred_excess', 'stat', 'stat_null', 'counts', 'success']
          ref. model             : pl
          n_sigma                : 1
          n_sigma_ul             : 2
          sqrt_ts_threshold_ul   : 2
          sed type init          : likelihood
    """

    tag = "FluxPointsEstimator"

    def __init__(
        self,
        energy_edges=[1, 10] * u.TeV,
        sum_over_energy_groups=False,
        n_jobs=None,
        parallel_backend=None,
        allow_multiple_telescopes=False,
        **kwargs,
    ):
        self.energy_edges = energy_edges
        self.sum_over_energy_groups = sum_over_energy_groups
        self.n_jobs = n_jobs
        self.parallel_backend = parallel_backend
        self.allow_multiple_telescopes = allow_multiple_telescopes

        fit = Fit(confidence_opts={"backend": "scipy"})
        kwargs.setdefault("fit", fit)
        super().__init__(**kwargs)

    def run(self, datasets):
        """Run the flux point estimator for all energy groups.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        flux_points : `FluxPoints`
            Estimated flux points.
        """
        if not isinstance(datasets, DatasetsActor):
            datasets = Datasets(datasets=datasets)

        if not datasets.energy_axes_are_aligned:
            raise ValueError("All datasets must have aligned energy axes.")

        telescopes = []
        for d in datasets:
            if d.meta_table is not None and "TELESCOP" in d.meta_table.colnames:
                telescopes.extend(list(d.meta_table["TELESCOP"].flatten()))
        if len(np.unique(telescopes)) > 1 and not self.allow_multiple_telescopes:
            raise ValueError(
                "All datasets must use the same value of the 'TELESCOP' meta keyword."
            )

        meta = {
            "n_sigma": self.n_sigma,
            "n_sigma_ul": self.n_sigma_ul,
            "sed_type_init": "likelihood",
        }

        rows = parallel.run_multiprocessing(
            self.estimate_flux_point,
            zip(
                repeat(datasets),
                self.energy_edges[:-1],
                self.energy_edges[1:],
            ),
            backend=self.parallel_backend,
            pool_kwargs=dict(processes=self.n_jobs),
            task_name="Energy bins",
        )

        table = Table(rows, meta=meta)
        model = _get_reference_model(datasets.models[self.source], self.energy_edges)
        return FluxPoints.from_table(
            table=table,
            reference_model=model.copy(),
            gti=datasets.gti,
            format="gadf-sed",
        )

    def estimate_flux_point(self, datasets, energy_min, energy_max):
        """Estimate flux point for a single energy group.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        energy_min, energy_max : `~astropy.units.Quantity`
            Energy bounds to compute the flux point for.

        Returns
        -------
        result : dict
            Dictionary with results for the flux point.
        """
        datasets_sliced = datasets.slice_by_energy(
            energy_min=energy_min, energy_max=energy_max
        )
        if self.sum_over_energy_groups:
            datasets_sliced = datasets_sliced.__class__(
                [_.to_image(name=_.name) for _ in datasets_sliced]
            )

        if len(datasets_sliced) > 0:
            if datasets.models is not None:
                models_sliced = datasets.models._slice_by_energy(
                    energy_min=energy_min,
                    energy_max=energy_max,
                    sum_over_energy_groups=self.sum_over_energy_groups,
                )
                datasets_sliced.models = models_sliced

            return super().run(datasets=datasets_sliced)
        else:
            log.warning(f"No dataset contribute in range {energy_min}-{energy_max}")
            model = _get_reference_model(
                datasets.models[self.source], self.energy_edges
            )
            return self._nan_result(datasets, model, energy_min, energy_max)

    def _nan_result(self, datasets, model, energy_min, energy_max):
        """NaN result."""
        energy_axis = MapAxis.from_energy_edges([energy_min, energy_max])

        with np.errstate(invalid="ignore", divide="ignore"):
            result = model.reference_fluxes(energy_axis=energy_axis)
            # convert to scalar values
            result = {key: value.item() for key, value in result.items()}

        result.update(
            {
                "norm": np.nan,
                "stat": np.nan,
                "success": False,
                "norm_err": np.nan,
                "ts": np.nan,
                "counts": np.zeros(len(datasets)),
                "npred": np.nan * np.zeros(len(datasets)),
                "npred_excess": np.nan * np.zeros(len(datasets)),
                "datasets": datasets.names,
            }
        )

        if "errn-errp" in self.selection_optional:
            result.update({"norm_errp": np.nan, "norm_errn": np.nan})

        if "ul" in self.selection_optional:
            result.update({"norm_ul": np.nan})

        if "scan" in self.selection_optional:
            norm_scan = self.norm.copy().scan_values
            result.update({"norm_scan": norm_scan, "stat_scan": np.nan * norm_scan})

        if "sensitivity" in self.selection_optional:
            result.update({"norm_sensitivity": np.nan})

        return result


class FluxCollectionEstimator:
    """Estimate the flux points from a collection of sources simultaneously.

    This estimator computes spectral flux points for a *collection of sources*
    over a set of predefined energy bins. The normalizations of all
    `~gammapy.modeling.models.SkyModel` objects listed in ``models`` are fitted
    jointly in each energy bin, while all other `~gammapy.modeling.models.SkyModel`
    components in the datasets remain frozen. Re-optimization of each dataset’s
    background model is optional.

    The operation can be performed either with the standard likelihood optimizer
    (`~gammapy.modeling.Fit`) or with a sampler
    (`~gammapy.modeling.Sampler`), which can be used to derive asymmetric
    errors and upper limits.

    By default, this requires the ultranest package to be installed.

    Parameters
    ----------
    energy_edges : `~astropy.units.Quantity`
        Energy edges of the flux point bins.
    models : `~gammapy.modeling.Models` or list
        Source models for which the flux points are computed (others are frozen).
    n_sigma : float, optional
        Number of sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Number of sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    norm : `~gammapy.modeling.Parameter`, optional
        Norm parameter used for the fit.
        Default is None and a new parameter is created with value=1, name="norm".
        If the `solver` is a sampler the default prior is uniform between [-10, 10].
    solver : `~gammapy.modeling.Fit` or `~gammapy.modeling.Sampler`
        Fit or Sampler instance specifying the backend and options.
        Default is a Sampler with options live_points=300, frac_remain=0.3.
    reoptimize : bool
        Whether to reoptimize each dataset.background_model. Default is False.
        Only SkyModel given in `models` will be fitted the others remain frozen,
        regardless of this option.
    selection_optional : list of str, optional
        Which additional quantities to estimate. Available options are:
            * "errn-errp": estimate asymmetric errors on flux.
        Fit solver computes upper limits if sqrt(TS) < n_sigma_ul.
        Sampler solver always compute errn-errp and ul.
    """

    def __init__(
        self,
        energy_edges,
        models,
        n_sigma=1,
        n_sigma_ul=2,
        norm=None,
        solver=None,
        reoptimize=False,
        selection_optional=None,
    ):
        self.n_sigma = n_sigma
        self.n_sigma_ul = n_sigma_ul

        self.models = models
        self.ns = len(models)

        self.solver = solver if solver is not None else self._default_solver
        self.reoptimize = reoptimize

        if selection_optional is None:
            selection_optional = []
        self.selection_optional = selection_optional

        self.norm = norm if norm is not None else self._default_norm

        self.energy_edges = energy_edges

        self.energy_unit = "TeV"
        self.dnde_unit = u.Unit("cm-2 s-1 TeV-1")

    @property
    def _default_norm(self):
        prior = None
        if isinstance(self.solver, Sampler):
            prior = UniformPrior(min=-10, max=10)
        return Parameter(name="norm", value=1, unit="", prior=prior)

    @property
    def _default_solver(self):
        """Return an ultranest Sampler with options live_points=300, frac_remain=0.3."""
        return Sampler(
            backend="ultranest",
            sampler_opts={"live_points": 300, "frac_remain": 0.3},
        )

    @property
    def _available_keys(self):
        # TODO: what about 'npred' key
        keys = ["norm", "norm_ul", "ts"]
        if isinstance(self.solver, Fit):
            keys.append("norm_err")
        if isinstance(self.solver, Sampler) or "errn-errp" in self.selection_optional:
            keys.extend(["norm_errn", "norm_errp"])
        return keys

    def _empty_result_dict(self):
        """Build an empty dictionary to store results."""
        n_sources = len(self.models)
        empty_fp_result = {key: np.zeros(n_sources) for key in self._available_keys}
        empty_fp_result["npred"] = np.zeros(n_sources)
        return empty_fp_result

    @property
    def _energy_axis(self):
        """Energy axis for the input edges."""
        return MapAxis.from_energy_edges(self.energy_edges, name="energy", interp="log")

    def _prepare_dataset(self, dataset, spectral_norm_models):
        """Build NPredTemplateModel for a given dataset.

        Create one single template summing all frozen sources and create one template per free source with free norm.
        """
        # TODO: can we avoid the deep copy? Does is remove the cached evaluators?
        fp_dataset = dataset.copy(name=dataset.name)

        source_names = set(Models(self.models).names)
        frozen_names = set(dataset._evaluators.keys()) - source_names

        fp_models = []
        npred_free = dataset.npred_signal(source_names, stack=False)
        for idx, npred in enumerate(npred_free.split_by_axis("models")):
            name = npred_free.geom.axes["models"].center[idx]
            template_model = TemplateNPredModel(
                npred,
                name=name + "_" + fp_dataset.name,
                spectral_model=spectral_norm_models[name],
                datasets_names=[fp_dataset.name],
            )
            fp_models.append(template_model)

        npred_frozen = fp_dataset.npred_signal(frozen_names, stack=True)
        bkg_frozen = TemplateNPredModel(
            npred_frozen,
            name="frozen_" + fp_dataset.name,
            datasets_names=[fp_dataset.name],
        )
        bkg_frozen.spectral_model.norm.frozen = True

        bkg_model = self._get_bkg(fp_dataset)
        fp_dataset.models = Models(fp_models + [bkg_frozen] + bkg_model)

        return fp_dataset

    def _prepare_datasets(self, datasets):
        """Define datasets with cached npred models to be renormalized."""
        norm_models = {}
        for m in self.models:
            norm_models[m.name] = PowerLawNormSpectralModel(norm=self.norm.copy())
            norm_models[m.name].tilt.frozen = True

        fp_datasets = [self._prepare_dataset(d, norm_models) for d in datasets]
        return Datasets(fp_datasets), norm_models

    def _get_bkg(self, d):
        if d.background_model:
            bkg_model = [d.background_model.copy(name=d.background_model.name)]
            if not self.reoptimize:
                bkg_model[0].freeze()
        else:
            bkg_model = []
        return bkg_model

    @staticmethod
    def _compute_npred(datasets, param, model):
        """Compute npred within the datasets masks."""
        npred = 0
        for kd, d in enumerate(datasets):
            name = model.name + "_" + d.name
            if d.evaluators[name].contributes:
                npred_map = Map.from_geom(d.counts.geom)
                npred_map.stack(d.evaluators[name].compute_npred())
                npred += np.nansum(npred_map.data * d.mask.data) * param.value
        return npred

    @staticmethod
    def _compute_ts(datasets, param):
        """Test statistic against no source as null hypothesis."""
        cash = datasets._stat_sum_likelihood()
        with Parameters([param]).restore_status():
            param.value = 0
            cash0 = datasets._stat_sum_likelihood()
        return cash0 - cash

    def _run_fit(self, fp_datasets, spectral_models):
        """Compute flux estimates, uncertainties and UL using log-likelihood profile (i.e. using `~gammapy.modeling.Fit`)."""
        fit_results = self.solver.run(fp_datasets)

        fp_result = self._empty_result_dict()
        fp_result["solver_results"] = fit_results

        for km, (m, spec) in enumerate(zip(self.models, spectral_models.values())):
            norm_param = spec.norm
            norm = norm_param.value
            error = norm_param.error

            npred = self._compute_npred(fp_datasets, norm_param, m)
            fp_result["npred"][km] = npred

            fp_result["norm"][km] = norm

            fp_result["norm_err"][km] = error
            if "errn-errp" in self.selection_optional:
                res = self.solver.confidence(
                    datasets=fp_datasets,
                    parameter=norm_param,
                    sigma=self.n_sigma,
                )
                fp_result["norm_errn"][km] = res["errn"]
                fp_result["norm_errp"][km] = res["errp"]

            ts_null = self._compute_ts(fp_datasets, norm_param)
            fp_result["ts"][km] = ts_null

            if np.sign(ts_null) * np.sqrt(np.abs(ts_null)) < self.n_sigma_ul:
                res = self.solver.confidence(
                    datasets=fp_datasets,
                    parameter=norm_param,
                    sigma=self.n_sigma_ul,
                )
                fp_result["norm_ul"][km] = norm + res["errp"]
            else:
                fp_result["norm_ul"][km] = np.nan
        return fp_result

    def _run_sampler(self, fp_datasets, spectral_models):
        """Compute npred, dnde, TS, errn, errp, and ul."""
        sampler_results = self.solver.run(fp_datasets).sampler_results

        fp_result = self._empty_result_dict()
        fp_result["solver_results"] = sampler_results

        points = sampler_results["weighted_samples"]["points"]
        weights = sampler_results["weighted_samples"]["weights"]

        quantiles = {
            "norm": 50,
            "norm_errn": 100 * stats.norm.cdf(-self.n_sigma),
            "norm_errp": 100 * stats.norm.cdf(self.n_sigma),
            "norm_ul": 100 * stats.norm.cdf(self.n_sigma_ul),
        }

        for km, (m, spec) in enumerate(zip(self.models, spectral_models.values())):
            samples = points[:, km]
            norm_param = spec.norm

            percentiles = {
                k: np.percentile(samples, q, weights=weights, method="inverted_cdf")
                for k, q in quantiles.items()
            }

            norm = percentiles["norm"]
            norm_param.value = norm  # set before TS computation below

            fp_result["norm"][km] = norm
            fp_result["norm_errn"][km] = norm - percentiles["norm_errn"]
            fp_result["norm_errp"][km] = percentiles["norm_errp"] - norm
            fp_result["norm_ul"][km] = percentiles["norm_ul"]
            fp_result["npred"][km] = self._compute_npred(fp_datasets, norm_param, m)

        # compute TS after norm value is set to median for all models
        for km, spec in enumerate(spectral_models.values()):
            norm_param = spec.norm
            ts_null = self._compute_ts(fp_datasets, norm_param)
            fp_result["ts"][km] = ts_null

        return fp_result

    def run(self, datasets):
        """Compute flux point in each energy band.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets used to compute the flux points. They must share the same geometry.

        Returns
        -------
        result : dict
            Dict with results
        """
        datasets = Datasets(datasets)
        if len(datasets) == 0:
            raise ValueError("datasets cannot be empty")

        if not datasets.is_all_same_geom:
            raise ValueError("Inconsistent geometries between datasets")

        for d in datasets:
            d.npred()  # precompute npred

        fp_datasets, spectral_models = self._prepare_datasets(datasets)

        fp_results = []

        for ke, (emin, emax) in enumerate(self._energy_axis.iter_by_edges):
            with set_and_restore_mask_fit(
                fp_datasets, energy_min=emin, energy_max=emax
            ):
                if isinstance(self.solver, Sampler):
                    fp_result = self._run_sampler(fp_datasets, spectral_models)
                else:
                    fp_result = self._run_fit(fp_datasets, spectral_models)

            fp_results.append(fp_result)

        return self._get_flux_points_dict(fp_results)

    def _get_flux_points_dict(self, fp_results):
        """Extract flux points for each model from list of results.

        Parameters
        ----------
        fp_results : dict
            dict used to generate the flux points table.

        Returns
        -------
        result : dict
            Dictionary with results.
        """

        def build_fp_from_idx(idx, model):
            table = Table()
            table["e_min"] = self._energy_axis.edges_min.to(self.energy_unit)
            table["e_max"] = self._energy_axis.edges_max.to(self.energy_unit)
            table["e_ref"] = self._energy_axis.center.to(self.energy_unit)
            table["ref_dnde"] = model(table["e_ref"]).to(self.dnde_unit)

            for key in self._available_keys:
                table[key] = np.array([fp[key][idx] for fp in fp_results])

            table.meta["SED_TYPE"] = "likelihood"

            return FluxPoints.from_table(
                table, reference_model=model.copy(), format="gadf-sed"
            )

        fp_dict = dict(
            energy_edges=self.energy_edges,
            solver_results=np.array(
                [fp["solver_results"] for fp in fp_results], dtype=object
            ),
            flux_points={},
        )

        for idx, m in enumerate(self.models):
            model = _get_reference_model(m, self.energy_edges)
            fp_dict["flux_points"][m.name] = build_fp_from_idx(idx, model)

        if isinstance(self.solver, Sampler):
            weights = [
                fp["solver_results"]["weighted_samples"]["weights"] for fp in fp_results
            ]
            dnde_dict = {}
            for model_idx, m in enumerate(self.models):
                dnde_dict[m.name] = []
                dnde_ref = fp_dict["flux_points"][m.name]["dnde_ref"].squeeze()
                for dnde, fp in zip(dnde_ref, fp_results):
                    points = fp["solver_results"]["weighted_samples"]["points"]
                    dnde_dict[m.name].append(dnde * points[:, model_idx])
            fp_dict["samples"] = dict(dnde=dnde_dict, weights=weights)

        return fp_dict



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import astropy.io.fits as fits
from astropy.table import Table
from astropy.time import Time
from astropy import units as u
import pytest

from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data
from ..io import EventListReader, EventListWriter, ObservationTableReader


@requires_data()
def test_eventlist_reader_no_format():
    hess_events = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    fermi_events = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz"

    events = EventListReader().read(hess_events, format=None)
    assert len(events.table) == 11243

    events = EventListReader().read(fermi_events, format=None)
    assert len(events.table) == 32843


@requires_data()
def test_eventlist_reader_unkwnown_format(tmpdir):
    hess_events = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )

    with fits.open(hess_events) as hdulist:
        hdu_events = hdulist["EVENTS"]
        hdu_events.header["HDUCLASS"] = "bad"

        hdulist.writeto(tmpdir / "tmp.fits")

    with pytest.raises(ValueError):
        EventListReader().read(tmpdir / "tmp.fits", format=None)


def test_eventlist_writer_unkwnown_format():
    with pytest.raises(ValueError):
        EventListWriter().to_hdu("tmp.fits", format="unknown")


@requires_data()
def test_eventlist_reader_empty_gadf_table():
    swgo_events = "$GAMMAPY_DATA/tests/format/swgo/map_irfs/DummyEvents.fits.gz"

    with pytest.raises(ValueError) as err:
        EventListReader().read(swgo_events, format="gadf")
    assert "ENERGY" in str(err.value)
    assert "TIME" in str(err.value)
    assert "RA" in str(err.value)
    assert "DEC" in str(err.value)


@requires_data()
def test_observationtable_reader_unknown_specified_format(tmpdir):
    hess_obs_table = make_path("$GAMMAPY_DATA/hess-dl3-dr1/obs-index.fits.gz")

    with fits.open(hess_obs_table) as hdulist:
        hdu_obs_table = hdulist["OBS_INDEX"]
        hdu_obs_table.header["HDUCLASS"] = "unknown-format"

        hdulist.writeto(tmpdir / "tmp.fits")

    with pytest.raises(ValueError):
        ObservationTableReader().read(tmpdir / "tmp.fits")


@requires_data()
def test_observationtable_reader_unspecified_format(tmpdir, caplog):
    hess_obs_table = make_path("$GAMMAPY_DATA/hess-dl3-dr1/obs-index.fits.gz")

    with fits.open(hess_obs_table) as hdulist:
        hdu_obs_table = hdulist["OBS_INDEX"]
        hdu_obs_table.header.remove("HDUCLASS")

        filename = "tmp.fits"
        hdulist.writeto(tmpdir / filename)

        ObservationTableReader().read(tmpdir / "tmp.fits")
        assert (
            f"Could not infer the file format from metadata in {tmpdir / filename}, assuming GADF."
            in [_.message for _ in caplog.records]
        )


def test_observationtable_reader_gadf_converter_mandatory_keywords():
    t = Table({"RA_PNT": [1.0]}, units={"RA_PNT": u.deg})
    with pytest.raises(RuntimeError):
        ObservationTableReader._from_gadf_table(t)


def test_observationtable_reader_gadf_converter_missing_time_keywords(caplog):
    t = Table({"OBS_ID": ["1"], "TSTART": [Time("2012-01-01T00:30:00")]}, meta={})

    obs_table = ObservationTableReader._from_gadf_table(t)
    assert (
        "Found column TSTART or TSTOP in GADF table, but can not create columns in internal format due to missing header keywords in file."
        in [_.message for _ in caplog.records]
    )
    assert obs_table.keys() == ["OBS_ID"]


def test_observationtable_reader_gadf_converter_time_conversion():
    t = Table(
        {"OBS_ID": ["1"], "TSTART": [100]},
        meta={
            "MJDREFI": 50000,
            "MJDREFF": 100.0,
            "TIMEUNIT": "s",
        },
    )
    obs_table = ObservationTableReader._from_gadf_table(t)
    assert obs_table.keys() == ["OBS_ID", "TSTART"]
    assert isinstance(obs_table["TSTART"], Time)


def test_observationtable_reader_gadf_converter_time_unit_not_understood(caplog):
    t = Table(
        {"OBS_ID": ["1"], "TSTART": [100]},
        meta={
            "MJDREFI": 50000,
            "MJDREFF": 100.0,
            "TIMEUNIT": "-",
        },
    )
    obs_table = ObservationTableReader._from_gadf_table(t)
    assert (
        "Unit for column TSTART not understood. Unit with dimension of time expected."
        in [_.message for _ in caplog.records]
    )
    assert obs_table.keys() == ["OBS_ID"]


def test_observationtable_reader_gadf_converter_time_unit_wrong_dimension(caplog):
    t = Table(
        {"OBS_ID": ["1"], "TSTART": [100]},
        meta={
            "MJDREFI": 50000,
            "MJDREFF": 100.0,
            "TIMEUNIT": "m",
        },
    )
    obs_table = ObservationTableReader._from_gadf_table(t)
    assert (
        "Unit for column TSTART is in wrong dimension. Unit with dimension of time expected."
        in [_.message for _ in caplog.records]
    )
    assert obs_table.keys() == ["OBS_ID"]


def test_observationtable_reader_gadf_converter_invalid_time_datatype(caplog):
    t = Table(
        {"OBS_ID": ["1"], "TSTOP": ["not_a_number"]},
        meta={
            "MJDREFI": 50000,
            "MJDREFF": 100.0,
            "TIMEUNIT": "s",
        },
    )
    obs_table = ObservationTableReader._from_gadf_table(t)
    assert "Could not convert type for column TSTOP to float64." in [
        _.message for _ in caplog.records
    ]
    assert obs_table.keys() == ["OBS_ID"]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from astropy.coordinates import Angle
from astropy.nddata import NoOverlapError
import gammapy.utils.parallel as parallel
from gammapy.datasets import Datasets, MapDataset, MapDatasetOnOff, SpectrumDataset
from .core import Maker
from .safe import SafeMaskMaker

log = logging.getLogger(__name__)


__all__ = [
    "DatasetsMaker",
]


class DatasetsMaker(Maker, parallel.ParallelMixin):
    """Run makers in a chain.

    Parameters
    ----------
    makers : list of `Maker` objects
        Makers.
    stack_datasets : bool, optional
        If True, stack into the reference dataset (see `run` method arguments).
        Default is True.
    n_jobs : int, optional
        Number of processes to run in parallel.
        Default is one, unless `~gammapy.utils.parallel.N_JOBS_DEFAULT` was modified.
    cutout_mode : {'trim', 'partial', 'strict'}
        Used only to cutout the reference `MapDataset` around each processed observation.
        Mode is an option for Cutout2D, for details see `~astropy.nddata.utils.Cutout2D`.
        Default is "trim".
    cutout_width : tuple of `~astropy.coordinates.Angle`, optional
        Angular sizes of the region in (lon, lat) in that specific order.
        If only one value is passed, a square region is extracted.
        If None it returns an error, except if the list of makers includes a `SafeMaskMaker`
        with the offset-max method defined. In that case it is set to two times `offset_max`.
        Default is None.
    parallel_backend : {'multiprocessing', 'ray'}, optional
        Which backend to use for multiprocessing.
        Default is None.
    """

    tag = "DatasetsMaker"

    def __init__(
        self,
        makers,
        stack_datasets=True,
        n_jobs=None,
        cutout_mode="trim",
        cutout_width=None,
        parallel_backend=None,
    ):
        self.log = logging.getLogger(__name__)
        self.makers = makers
        self.cutout_mode = cutout_mode

        if cutout_width is not None:
            cutout_width = Angle(cutout_width)

        self.cutout_width = cutout_width
        self._apply_cutout = True

        if self.cutout_width is None:
            if self.offset_max is None:
                self._apply_cutout = False
            else:
                self.cutout_width = 2 * self.offset_max

        self.n_jobs = n_jobs
        self.parallel_backend = parallel_backend
        self.stack_datasets = stack_datasets

        self._datasets = []
        self._error = False

    @property
    def offset_max(self):
        maker = self.safe_mask_maker
        if maker is not None and hasattr(maker, "offset_max"):
            return maker.offset_max

    @property
    def safe_mask_maker(self):
        for m in self.makers:
            if isinstance(m, SafeMaskMaker):
                return m

    def make_dataset(self, dataset, observation):
        """Make single dataset.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Reference dataset.
        observation : `Observation`
            Observation.
        """
        obs_point = observation.get_pointing_icrs(observation.tmid)
        if self._apply_cutout:
            cutouts_kwargs = {
                "position": obs_point.galactic,
                "width": self.cutout_width,
                "mode": self.cutout_mode,
            }
            try:
                dataset_obs = dataset.cutout(
                    **cutouts_kwargs,
                )
            except NoOverlapError:
                log.warning(
                    f"{observation.obs_id} discarded because"
                    " it is outside the FOV of the reference geom."
                )
                return None

        else:
            dataset_obs = dataset.copy()

        if dataset.models is not None:
            models = dataset.models.copy()
            models.reassign(dataset.name, dataset_obs.name)
            dataset_obs.models = models

        log.info(f"Computing dataset for observation {observation.obs_id}")

        for maker in self.makers:
            log.info(f"Running {maker.tag}")
            dataset_obs = maker.run(dataset=dataset_obs, observation=observation)

        return dataset_obs

    def callback(self, dataset):
        if self.stack_datasets and dataset is not None:
            if type(self._dataset) is MapDataset and type(dataset) is MapDatasetOnOff:
                dataset = dataset.to_map_dataset(name=dataset.name)
            self._dataset.stack(dataset)
        else:
            self._datasets.append(dataset)

    def error_callback(self, dataset):
        # parallel run could cause a memory error with non-explicit message.
        self._error = True

    def run(self, dataset, observations, datasets=None):
        """Run data reduction.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Reference dataset (used only for stacking if datasets are provided).
        observations : `Observations`
            Observations.
        datasets : `~gammapy.datasets.Datasets`
            Base datasets, if provided its length must be the same as the observations.

        Returns
        -------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        """
        if isinstance(dataset, MapDataset):
            # also valid for Spectrum as it inherits from MapDataset
            self._dataset = dataset
        else:
            raise TypeError("Invalid reference dataset.")

        if isinstance(dataset, SpectrumDataset):
            self._apply_cutout = False

        if datasets is not None:
            self._apply_cutout = False
        else:
            datasets = len(observations) * [dataset]

        n_jobs = min(self.n_jobs, len(observations))

        parallel.run_multiprocessing(
            self.make_dataset,
            zip(datasets, observations),
            backend=self.parallel_backend,
            pool_kwargs=dict(processes=n_jobs),
            method="apply_async",
            method_kwargs=dict(
                callback=self.callback,
                error_callback=self.error_callback,
            ),
            task_name="Data reduction",
        )

        if self._error:
            raise RuntimeError("Execution of a sub-process failed")

        if self.stack_datasets:
            return Datasets([self._dataset])

        valid_datasets = [d for d in self._datasets if d is not None]
        lookup = {
            d.meta_table["OBS_ID"][0]: idx for idx, d in enumerate(valid_datasets)
        }
        return Datasets(
            [
                valid_datasets[lookup[obs.obs_id]]
                for obs in observations
                if obs.obs_id in lookup
            ]
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import copy
import logging
from gammapy.utils.cache import cachemethod
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.io import fits
from astropy.table import QTable, Table
from astropy.utils import lazyproperty
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs.utils import (
    proj_plane_pixel_area,
    proj_plane_pixel_scales,
    wcs_to_celestial_frame,
)
from regions import (
    CompoundSkyRegion,
    PixCoord,
    PointSkyRegion,
    RectanglePixelRegion,
    Regions,
    SkyRegion,
)
import matplotlib.pyplot as plt
from gammapy.utils.regions import (
    compound_region_center,
    compound_region_to_regions,
    regions_to_compound_region,
)
from gammapy.visualization.utils import ARTIST_TO_LINE_PROPERTIES
from ..axes import MapAxes
from ..coord import MapCoord
from ..core import Map
from ..geom import Geom, pix_tuple_to_idx
from ..utils import _check_width
from ..wcs import WcsGeom

log = logging.getLogger(__name__)


__all__ = ["RegionGeom"]


def _parse_regions(regions):
    if isinstance(regions, str):
        regions = Regions.parse(data=regions, format="ds9")
    elif isinstance(regions, SkyRegion):
        regions = [regions]
    elif isinstance(regions, SkyCoord):
        regions = [PointSkyRegion(center=regions)]
    elif isinstance(regions, list) and len(regions) == 0:
        regions = None
    return regions


class RegionGeom(Geom):
    """Map geometry representing a region on the sky.

    The spatial component of the geometry is made up of a
    single pixel with an arbitrary shape and size. It can
    also have any number of non-spatial dimensions. This class
    represents the geometry for the `RegionNDMap` maps.

    Parameters
    ----------
    region : `~regions.SkyRegion`
        Region object.
    axes : list of `MapAxis`
        Non-spatial data axes.
    wcs : `~astropy.wcs.WCS`
        Optional WCS object to project the region if needed.
    binsz_wcs : `~astropy.units.Quantity`
        Angular bin size of the underlying `~WcsGeom` used to evaluate
        quantities in the region. Default is "0.1 deg". This default
        value is adequate for the majority of use cases. If a WCS object
        is provided, the input of ``binsz_wcs`` is overridden.

    Notes
    -----
    - For further explanation and examples, see :doc:`/user-guide/maps/regionmap`

    """

    is_regular = True
    is_allsky = False
    is_hpx = False
    is_region = True

    _slice_spatial_axes = slice(0, 2)
    _slice_non_spatial_axes = slice(2, None)
    projection = "TAN"

    def __init__(self, region, axes=None, wcs=None, binsz_wcs="0.1 deg"):
        self._region = region
        self._axes = MapAxes.from_default(axes, n_spatial_axes=2)
        self._binsz_wcs = u.Quantity(binsz_wcs)

        if wcs is None and region is not None:
            if isinstance(region, CompoundSkyRegion):
                self._center = compound_region_center(region)
            else:
                self._center = region.center

            wcs = WcsGeom.create(
                binsz=binsz_wcs,
                skydir=self._center,
                proj=self.projection,
                frame=self._center.frame.name,
            ).wcs
            self._wcs = wcs
            # TODO : can we get the width before defining the wcs ?
            wcs = WcsGeom.create(
                binsz=binsz_wcs,
                width=tuple(self.width),
                skydir=self._center,
                proj=self.projection,
                frame=self._center.frame.name,
            ).wcs

        self._wcs = wcs
        self.ndim = len(self.data_shape)

    @property
    def frame(self):
        """Coordinate system, either Galactic ("galactic") or Equatorial ("icrs")."""
        if self.region is None:
            return "icrs"
        try:
            return self.region.center.frame.name
        except AttributeError:
            return wcs_to_celestial_frame(self.wcs).name

    @property
    def binsz_wcs(self):
        """Angular bin size of the underlying `~WcsGeom`.

        Returns
        -------
        binsz_wcs: `~astropy.coordinates.Angle`
            Angular bin size.
        """
        return Angle(proj_plane_pixel_scales(self.wcs), unit="deg")

    @lazyproperty
    def _rectangle_bbox(self):
        if self.region is None:
            raise ValueError("Region definition required.")

        regions = compound_region_to_regions(self.region)
        regions_pix = [_.to_pixel(self.wcs) for _ in regions]

        bbox = regions_pix[0].bounding_box

        for region_pix in regions_pix[1:]:
            bbox = bbox.union(region_pix.bounding_box)

        try:
            rectangle_pix = bbox.to_region()
        except ValueError:
            rectangle_pix = RectanglePixelRegion(
                center=PixCoord(*bbox.center[::-1]), width=1, height=1
            )
        return rectangle_pix.to_sky(self.wcs)

    @property
    def width(self):
        """Width of bounding box of the region.

        Returns
        -------
        width : `~astropy.units.Quantity`
            Dimensions of the region in both spatial dimensions.
            Units: ``deg``
        """
        rectangle = self._rectangle_bbox
        return u.Quantity([rectangle.width.to("deg"), rectangle.height.to("deg")])

    @property
    def region(self):
        """The spatial component of the region geometry as a `~regions.SkyRegion`."""
        return self._region

    @property
    def is_all_point_sky_regions(self):
        """Whether regions are all point regions."""
        regions = compound_region_to_regions(self.region)
        return np.all([isinstance(_, PointSkyRegion) for _ in regions])

    @property
    def axes(self):
        """List of non-spatial axes."""
        return self._axes

    @property
    def axes_names(self):
        """All axes names."""
        return ["lon", "lat"] + self.axes.names

    @property
    def wcs(self):
        """WCS projection object."""
        return self._wcs

    @property
    def center_coord(self):
        """Center coordinate of the region as a `astropy.coordinates.SkyCoord`."""
        return self.pix_to_coord(self.center_pix)

    @property
    def center_pix(self):
        """Pixel values corresponding to the center of the region."""
        return tuple((np.array(self.data_shape) - 1.0) / 2)[::-1]

    @lazyproperty
    def center_skydir(self):
        """Sky coordinate of the center of the region."""
        if self.region is None:
            return SkyCoord(np.nan * u.deg, np.nan * u.deg)

        return self._rectangle_bbox.center

    @property
    def npix(self):
        """Number of spatial pixels."""
        return ([1], [1])

    def contains(self, coords):
        """Check if a given map coordinate is contained in the region.

        Requires the `.region` attribute to be set.
        For `PointSkyRegion` the method always returns True.

        Parameters
        ----------
        coords : tuple, dict, `MapCoord` or `~astropy.coordinates.SkyCoord`
            Object containing coordinate arrays we wish to check for inclusion
            in the region.

        Returns
        -------
        mask : `~numpy.ndarray`
            Boolean array with the same shape as the input that indicates
            which coordinates are inside the region.
        """
        if self.region is None:
            raise ValueError("Region definition required.")

        coords = MapCoord.create(coords, frame=self.frame, axis_names=self.axes.names)

        if self.is_all_point_sky_regions:
            return np.ones(coords.skycoord.shape, dtype=bool)

        return self.region.contains(coords.skycoord, self.wcs)

    def contains_wcs_pix(self, pix):
        """Check if a given WCS pixel coordinate is contained in the region.

        For `PointSkyRegion` the method always returns True.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel coordinates.

        Returns
        -------
        containment : `~numpy.ndarray`
            Boolean array.
        """
        if self.is_all_point_sky_regions:
            return np.ones(pix[0].shape, dtype=bool)

        region_pix = self.region.to_pixel(self.wcs)
        return region_pix.contains(PixCoord(pix[0], pix[1]))

    def separation(self, position):
        """Angular distance between the center of the region and the given position.

        Parameters
        ----------
        position : `astropy.coordinates.SkyCoord`
            Sky coordinate we want the angular distance to.

        Returns
        -------
        sep : `~astropy.coordinates.Angle`
            The on-sky separation between the given coordinate and the region center.
        """
        return self.center_skydir.separation(position)

    @property
    def data_shape(self):
        """Shape of the `~numpy.ndarray` matching this geometry."""
        return self._shape[::-1]

    @property
    def data_shape_axes(self):
        """Shape of data of the non-spatial axes and unit spatial axes."""
        return self.axes.shape[::-1] + (1, 1)

    @property
    def _shape(self):
        """Number of bins in each dimension.

        The spatial dimension is always (1, 1), as a
        `RegionGeom` is not pixelized further.
        """
        return tuple((1, 1) + self.axes.shape)

    def get_coord(self, mode="center", frame=None, sparse=False, axis_name=None):
        """Get map coordinates from the geometry.

        Parameters
        ----------
        mode : {'center', 'edges'}, optional
            Get center or edge coordinates for the non-spatial axes.
            Default is "center".
        frame : str or `~astropy.coordinates.Frame`, optional
            Coordinate frame. Default is None.
        sparse : bool, optional
            Compute sparse coordinates. Default is False.
        axis_name : str, optional
            If mode = "edges", the edges will be returned for this axis only.
            Default is None.

        Returns
        -------
        coord : `~MapCoord`
            Map coordinate object.
        """
        if mode == "edges" and axis_name is None:
            raise ValueError("Mode 'edges' requires axis name")

        coords = self.axes.get_coord(mode=mode, axis_name=axis_name)
        coords["skycoord"] = self.center_skydir.reshape((1, 1))

        if frame is None:
            frame = self.frame

        coords = MapCoord.create(coords, frame=self.frame).to_frame(frame)

        if not sparse:
            coords = coords.broadcasted

        return coords

    def _pad_spatial(self, pad_width):
        raise NotImplementedError("Spatial padding of `RegionGeom` not supported")

    def crop(self):
        raise NotImplementedError("Cropping of `RegionGeom` not supported")

    def solid_angle(self):
        """Get solid angle of the region.

        Returns
        -------
        angle : `~astropy.units.Quantity`
            Solid angle of the region in steradians.
            Units: ``sr``
        """
        if self.region is None:
            raise ValueError("Region definition required.")

        # compound regions do not implement area()
        # so we use the mask representation and estimate the area
        # from the pixels in the mask using oversampling
        if isinstance(self.region, CompoundSkyRegion):
            # oversample by a factor of ten
            oversampling = 10.0
            wcs = self.to_binsz_wcs(self.binsz_wcs / oversampling).wcs
            pixel_region = self.region.to_pixel(wcs)
            mask = pixel_region.to_mask()
            area = np.count_nonzero(mask) / oversampling**2
        else:
            # all other types of regions should implement area
            area = self.region.to_pixel(self.wcs).area

        solid_angle = area * proj_plane_pixel_area(self.wcs) * u.deg**2
        return solid_angle.to("sr")

    def bin_volume(self):
        """If the `RegionGeom` has a non-spatial axis, it returns the volume of the region.
        If not, it returns the solid angle size.

        Returns
        -------
        volume : `~astropy.units.Quantity`
            Volume of the region.
        """
        bin_volume = self.solid_angle() * np.ones(self.data_shape)

        for idx, ax in enumerate(self.axes):
            shape = self.ndim * [1]
            shape[-(idx + 3)] = -1
            bin_volume = bin_volume * ax.bin_width.reshape(tuple(shape))

        return bin_volume

    def to_wcs_geom(self, width_min=None):
        """Get the minimal equivalent geometry which contains the region.

        Parameters
        ----------
        width_min : `~astropy.quantity.Quantity`, optional
            Minimum width for the resulting geometry. Can be a single number or two,
            for different minimum widths in each spatial dimension.
            Default is None.

        Returns
        -------
        wcs_geom : `~WcsGeom`
            A WCS geometry object.
        """
        if width_min is not None:
            width = np.max(
                [self.width.to_value("deg"), _check_width(width_min)], axis=0
            )
        else:
            width = self.width
        wcs_geom_region = WcsGeom(wcs=self.wcs)
        wcs_geom = wcs_geom_region.cutout(position=self.center_skydir, width=width)
        wcs_geom = wcs_geom.to_cube(self.axes)
        return wcs_geom

    def to_binsz_wcs(self, binsz):
        """Change the bin size of the underlying WCS geometry.

        Parameters
        ----------
        binsz : float, str or `~astropy.quantity.Quantity`
            Bin size.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the same axes and region as the input,
            but different WCS pixelization.
        """
        new_geom = RegionGeom(self.region, axes=self.axes, binsz_wcs=binsz)
        return new_geom

    @cachemethod
    def get_wcs_coord_and_weights(self, factor=10):
        """Get the array of spatial coordinates and corresponding weights.

        The coordinates are the center of a pixel that intersects the region and
        the weights that represent which fraction of the pixel is contained
        in the region.

        Parameters
        ----------
        factor : int, optional
            Oversampling factor to compute the weights.
            Default is 10.

        Returns
        -------
        region_coord : `~MapCoord`
            MapCoord object with the coordinates inside
            the region.
        weights : `~numpy.ndarray`
            Weights representing the fraction of each pixel
            contained in the region.
        """
        wcs_geom = self.to_wcs_geom()

        weights = wcs_geom.to_image().region_weights(
            regions=[self.region], oversampling_factor=factor
        )

        mask = weights.data > 0
        weights = weights.data[mask]

        # Get coordinates
        coords = wcs_geom.get_coord(sparse=True).apply_mask(mask)
        return coords, weights

    def to_binsz(self, binsz):
        """Return self."""
        return self

    def to_cube(self, axes):
        """Append non-spatial axes to create a higher-dimensional geometry.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the added axes.
        """
        axes = copy.deepcopy(self.axes) + axes
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def to_image(self):
        """Remove non-spatial axes to create a 2D region.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry without any non-spatial axes.
        """
        return self._init_copy(region=self.region, wcs=self.wcs, axes=None)

    def upsample(self, factor, axis_name=None):
        """Upsample a non-spatial dimension of the region by a given factor.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the upsampled axis.
        """
        axes = self.axes.upsample(factor=factor, axis_name=axis_name)
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def downsample(self, factor, axis_name):
        """Downsample a non-spatial dimension of the region by a given factor.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the downsampled axis.
        """
        axes = self.axes.downsample(factor=factor, axis_name=axis_name)
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def pix_to_coord(self, pix):
        lon = np.where(
            (-0.5 < pix[0]) & (pix[0] < 0.5),
            self.center_skydir.data.lon,
            np.nan * u.deg,
        )
        lat = np.where(
            (-0.5 < pix[1]) & (pix[1] < 0.5),
            self.center_skydir.data.lat,
            np.nan * u.deg,
        )
        coords = (lon, lat)

        for p, ax in zip(pix[self._slice_non_spatial_axes], self.axes):
            coords += (ax.pix_to_coord(p),)

        return coords

    def pix_to_idx(self, pix, clip=False):
        idxs = list(pix_tuple_to_idx(pix))

        for i, idx in enumerate(idxs[self._slice_non_spatial_axes]):
            if clip:
                np.clip(idx, 0, self.axes[i].nbin - 1, out=idx)
            else:
                np.putmask(idx, (idx < 0) | (idx >= self.axes[i].nbin), -1)

        return tuple(idxs)

    def coord_to_pix(self, coords):
        # inherited docstring
        if isinstance(coords, tuple) and len(coords) == len(self.axes):
            skydir = self.center_skydir.transform_to(self.frame)
            coords = (skydir.data.lon, skydir.data.lat) + coords
        elif isinstance(coords, dict):
            valid_keys = ["lon", "lat", "skycoord"]
            if not any([_ in coords for _ in valid_keys]):
                coords.setdefault("skycoord", self.center_skydir)

        coords = MapCoord.create(coords, frame=self.frame, axis_names=self.axes.names)

        if self.region is None:
            pix = (0, 0)
        else:
            in_region = self.contains(coords.skycoord)

            x = np.zeros(coords.skycoord.shape)
            x[~in_region] = np.nan

            y = np.zeros(coords.skycoord.shape)
            y[~in_region] = np.nan

            pix = (x, y)

        pix += self.axes.coord_to_pix(coords)
        return pix

    def get_idx(self):
        idxs = [np.arange(n, dtype=float) for n in self.data_shape[::-1]]
        return np.meshgrid(*idxs[::-1], indexing="ij")[::-1]

    def _make_bands_cols(self):
        return []

    @classmethod
    def create(cls, region, **kwargs):
        """Create region geometry.

        The input region can be passed in the form of a ds9 string and will be parsed
        internally by `~regions.Regions.parse`. See:

        * https://astropy-regions.readthedocs.io/en/stable/region_io.html
        * http://ds9.si.edu/doc/ref/region.html

        Parameters
        ----------
        region : str or `~regions.SkyRegion`
            Region definition.
        **kwargs : dict
            Keyword arguments passed to `RegionGeom.__init__`.

        Returns
        -------
        geom : `RegionGeom`
            Region geometry.
        """
        return cls.from_regions(regions=region, **kwargs)

    def __str__(self):
        axes = ["lon", "lat"] + [_.name for _ in self.axes]
        try:
            frame = self.center_skydir.frame.name
            lon = self.center_skydir.data.lon.deg
            lat = self.center_skydir.data.lat.deg
        except AttributeError:
            frame, lon, lat = "", np.nan, np.nan

        return (
            f"{self.__class__.__name__}\n\n"
            f"\tregion     : {self.region.__class__.__name__}\n"
            f"\taxes       : {axes}\n"
            f"\tshape      : {self.data_shape[::-1]}\n"
            f"\tndim       : {self.ndim}\n"
            f"\tframe      : {frame}\n"
            f"\tcenter     : {lon:.1f} deg, {lat:.1f} deg\n"
        )

    def is_allclose(self, other, rtol_axes=1e-6, atol_axes=1e-6):
        """Compare two data IRFs for equivalency.

        Parameters
        ----------
        other : `RegionGeom`
            Region geometry to compare against.
        rtol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.
        atol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.

        Returns
        -------
        is_allclose : bool
            Whether the geometry is all close.
        """
        if not isinstance(other, self.__class__):
            return TypeError(f"Cannot compare {type(self)} and {type(other)}")

        if self.data_shape != other.data_shape:
            log.debug("RegionGeom data shape is not equal")
            return False

        axes_eq = self.axes.is_allclose(other.axes, rtol=rtol_axes, atol=atol_axes)
        if not axes_eq:
            log.debug("RegionGeom axes are not equal")
        # TODO: compare regions based on masks...
        regions_eq = True
        return axes_eq and regions_eq

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.is_allclose(other=other)

    def _to_region_table(self):
        """Export region to a FITS region table."""
        if self.region is None:
            raise ValueError("Region definition required.")

        region_list = compound_region_to_regions(self.region)

        pixel_region_list = []

        for reg in region_list:
            pixel_region_list.append(reg.to_pixel(self.wcs))

        table = Regions(pixel_region_list).serialize(format="fits")

        header = WcsGeom(wcs=self.wcs).to_header()
        table.meta.update(header)
        return table

    def to_hdulist(self, format="ogip", hdu_bands=None, hdu_region=None):
        """Convert geometry to HDU list.

        Parameters
        ----------
        format : {"ogip", "gadf", "ogip-sherpa"}
            HDU format. Default is "ogip".
        hdu_bands : str, optional
            Name or index of the HDU with the BANDS table.
            Default is None.
        hdu_region : str, optional
            Name or index of the HDU with the region table.
            Not used for the "gadf" format.
            Default is None.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            HDU list.

        """
        if hdu_bands is None:
            hdu_bands = "HDU_BANDS"
        if hdu_region is None:
            hdu_region = "HDU_REGION"
        if format != "gadf":
            hdu_region = "REGION"

        hdulist = fits.HDUList()

        hdulist.append(self.axes.to_table_hdu(hdu_bands=hdu_bands, format=format))

        # region HDU
        if self.region:
            region_table = self._to_region_table()

            region_hdu = fits.BinTableHDU(region_table, name=hdu_region)
            hdulist.append(region_hdu)

        return hdulist

    @classmethod
    def from_regions(cls, regions, **kwargs):
        """Create region geometry from list of regions.

        The regions are combined with union to a compound region.

        Note that this function is intended for combining close by
        regions into one `~gammapy.maps.RegionGeom`. Since it uses an
        underlying tangential projection, undefined behaviour may occur
        if the list of input regions contains regions far from each other.
        If required, one can explicitly pass an appropriate WCS, eg with a cartesian
        projection, through ``kwargs``.

        Parameters
        ----------
        regions : list of `~regions.SkyRegion` or str
            Regions.
        **kwargs: dict
            Keyword arguments forwarded to `RegionGeom`.

        Returns
        -------
        geom : `RegionGeom`
            Region map geometry.

        Examples
        --------
        >>> from gammapy.maps import RegionGeom, WcsGeom
        >>> from regions import CircleSkyRegion
        >>> from astropy.coordinates import SkyCoord
        >>> import astropy.units as u

        >>> list_region = [CircleSkyRegion(SkyCoord(326*u.deg, -13*u.deg), radius=0.4*u.deg),
        ... CircleSkyRegion(SkyCoord(325*u.deg, -14*u.deg), radius=0.3*u.deg)]
        >>> wcs = WcsGeom.create().wcs
        >>> region = RegionGeom.from_regions(list_region, wcs=wcs)
        >>> print(region)
        RegionGeom
        region     : CompoundSkyRegion
        axes       : ['lon', 'lat']
        shape      : (1, 1)
        ndim       : 2
        frame      : icrs
        center     : 325.5 deg, -13.5 deg

        """
        regions = _parse_regions(regions)

        if regions:
            regions = regions_to_compound_region(regions)

        return cls(region=regions, **kwargs)

    @classmethod
    def from_hdulist(cls, hdulist, format="ogip", hdu=None):
        """Read region table and convert it to region list.

        Parameters
        ----------
        hdulist : `~astropy.io.fits.HDUList`
            HDU list.
        format : {"ogip", "ogip-arf", "gadf"}
            HDU format. Default is "ogip".
        hdu : str, optional
            Name of the HDU. Default is None.

        Returns
        -------
        geom : `RegionGeom`
            Region map geometry.

        """
        region_hdu = "REGION"

        if format == "gadf" and hdu:
            region_hdu = f"{hdu}_{region_hdu}"

        if region_hdu in hdulist:
            try:
                region_table = QTable.read(hdulist[region_hdu])
                regions_pix = Regions.parse(data=region_table, format="fits")
            except TypeError:
                region_table = Table.read(hdulist[region_hdu])
                regions_pix = Regions.parse(data=region_table, format="fits")

            wcs = WcsGeom.from_header(region_table.meta).wcs
            regions = []

            for region_pix in regions_pix:
                # TODO: remove workaround once regions issue with fits serialization is sorted out
                # see https://github.com/astropy/regions/issues/400
                # requires update regions to >=v0.6
                region_pix.meta["include"] = True
                regions.append(region_pix.to_sky(wcs))

            region = regions_to_compound_region(regions)
        else:
            region, wcs = None, None

        if format == "ogip":
            hdu_bands = "EBOUNDS"
        elif format == "ogip-arf":
            hdu_bands = "SPECRESP"
        elif format == "gadf":
            hdu_bands = hdu + "_BANDS"
        else:
            raise ValueError(f"Unknown format {format}")

        axes = MapAxes.from_table_hdu(hdulist[hdu_bands], format=format)
        return cls(region=region, wcs=wcs, axes=axes)

    def union(self, other):
        """Stack a `RegionGeom` by making the union."""
        if not self == other:
            raise ValueError("Can only make union if extra axes are equivalent.")
        if other.region:
            if self.region:
                self._region = self.region.union(other.region)
            else:
                self._region = other.region

    def plot_region(self, ax=None, kwargs_point=None, path_effect=None, **kwargs):
        """Plot region in the sky.

        Parameters
        ----------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
            Axes to plot on. If no axes are given,
            the region is shown using the minimal
            equivalent WCS geometry.
            Default is None.
        kwargs_point : dict, optional
            Keyword arguments passed to `~matplotlib.lines.Line2D` for plotting
            of point sources. Default is None.
        path_effect : `~matplotlib.patheffects`, optional
            Path effect applied to artists and lines.
            Default is None.
        **kwargs : dict
            Keyword arguments forwarded to `~regions.PixelRegion.as_artist`.

        Returns
        -------
        ax : `~astropy.visualization.wcsaxes.WCSAxes`
            Axes to plot on.
        """
        if self.region:
            kwargs_point = kwargs_point or {}

            if ax is None:
                ax = plt.gca()

                if not isinstance(ax, WCSAxes):
                    ax.remove()
                    wcs_geom = self.to_wcs_geom()
                    m = Map.from_geom(geom=wcs_geom.to_image())
                    ax = m.plot(add_cbar=False, vmin=-1, vmax=0)

            kwargs.setdefault("facecolor", "None")
            user_color = kwargs.pop("color", kwargs.pop("c", None))
            kwargs.setdefault("edgecolor", kwargs.pop("ec", "C0"))
            if user_color is not None:
                kwargs["edgecolor"] = user_color

            kwargs_point.setdefault("marker", "*")

            for key, value in kwargs.items():
                key_point = ARTIST_TO_LINE_PROPERTIES.get(key, None)
                if key_point and key_point not in kwargs_point:
                    kwargs_point[key_point] = value

            if user_color is not None:
                kwargs_point.setdefault("color", user_color)
            if "color" in kwargs_point:
                kwargs_point["markeredgecolor"] = kwargs_point["color"]

            for region in compound_region_to_regions(self.region):
                region_pix = region.to_pixel(wcs=ax.wcs)

                if isinstance(region, PointSkyRegion):
                    artist = region_pix.as_artist(**kwargs_point)
                else:
                    artist = region_pix.as_artist(**kwargs)

                if path_effect:
                    artist.add_path_effect(path_effect)

                ax.add_artist(artist)

            return ax
        else:
            logging.info("Region definition required.")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from gammapy.utils.scripts import make_path
from gammapy.utils.metadata import CreatorMetaData
from .map import MapDataset, MapDatasetOnOff
from .utils import get_axes
from astropy.io import fits


__all__ = ["SpectrumDatasetOnOff", "SpectrumDataset"]

log = logging.getLogger(__name__)


class PlotMixin:
    """Plot mixin for the spectral datasets."""

    def plot_fit(
        self,
        ax_spectrum=None,
        ax_residuals=None,
        kwargs_spectrum=None,
        kwargs_residuals=None,
    ):
        """Plot spectrum and residuals in two panels.

        Calls `~SpectrumDataset.plot_excess` and `~SpectrumDataset.plot_residuals_spectral`.

        Parameters
        ----------
        ax_spectrum : `~matplotlib.axes.Axes`, optional
            Axes to plot spectrum on. Default is None.
        ax_residuals : `~matplotlib.axes.Axes`, optional
            Axes to plot residuals on. Default is None.
        kwargs_spectrum : dict, optional
            Keyword arguments passed to `~SpectrumDataset.plot_excess`. Default is None.
        kwargs_residuals : dict, optional
            Keyword arguments passed to `~SpectrumDataset.plot_residuals_spectral`. Default is None.

        Returns
        -------
        ax_spectrum, ax_residuals : `~matplotlib.axes.Axes`
            Spectrum and residuals plots.

        Examples
        --------
        >>> #Creating a spectral dataset
        >>> from gammapy.datasets import SpectrumDatasetOnOff
        >>> from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
        >>> filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
        >>> dataset = SpectrumDatasetOnOff.read(filename)
        >>> p = PowerLawSpectralModel()
        >>> dataset.models = SkyModel(spectral_model=p)
        >>> # optional configurations
        >>> kwargs_excess = {"color": "blue", "markersize":8, "marker":'s', }
        >>> kwargs_npred_signal = {"color": "black", "ls":"--"}
        >>> kwargs_spectrum = {"kwargs_excess":kwargs_excess, "kwargs_npred_signal":kwargs_npred_signal}
        >>> kwargs_residuals = {"color": "black", "markersize":4, "marker":'s', }
        >>> dataset.plot_fit(kwargs_residuals=kwargs_residuals, kwargs_spectrum=kwargs_spectrum)  # doctest: +SKIP
        """
        gs = GridSpec(7, 1)
        bool_visible_xticklabel = not (ax_spectrum is None and ax_residuals is None)
        ax_spectrum, ax_residuals = get_axes(
            ax_spectrum,
            ax_residuals,
            8,
            7,
            [gs[:5, :]],
            [gs[5:, :]],
            kwargs2={"sharex": ax_spectrum},
        )
        kwargs_spectrum = kwargs_spectrum or {}
        kwargs_residuals = kwargs_residuals or {}

        self.plot_excess(ax_spectrum, **kwargs_spectrum)

        self.plot_residuals_spectral(ax_residuals, **kwargs_residuals)

        method = kwargs_residuals.get("method", "diff")
        label = self._residuals_labels[method]
        ax_residuals.set_ylabel(f"Residuals\n{label}")
        plt.setp(ax_spectrum.get_xticklabels(), visible=bool_visible_xticklabel)
        self.plot_masks(ax=ax_spectrum)

        return ax_spectrum, ax_residuals

    def plot_counts(
        self, ax=None, kwargs_counts=None, kwargs_background=None, **kwargs
    ):
        """Plot counts and background.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axes to plot on. Default is None.
        kwargs_counts : dict, optional
            Keyword arguments passed to `~matplotlib.axes.Axes.hist` for the counts. Default is None.
        kwargs_background : dict, optional
            Keyword arguments passed to `~matplotlib.axes.Axes.hist` for the background. Default is None.
        **kwargs : dict, optional
            Keyword arguments passed to both `~matplotlib.axes.Axes.hist`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes object.
        """
        kwargs_counts = kwargs_counts or {}
        kwargs_background = kwargs_background or {}

        plot_kwargs = kwargs.copy()
        plot_kwargs.update(kwargs_counts)
        plot_kwargs.setdefault("label", "Counts")
        ax = self.counts.plot_hist(ax=ax, **plot_kwargs)

        plot_kwargs = kwargs.copy()
        plot_kwargs.update(kwargs_background)

        plot_kwargs.setdefault("label", "Background")
        self.background.plot_hist(ax=ax, **plot_kwargs)

        ax.legend(numpoints=1)
        return ax

    def plot_masks(self, ax=None, kwargs_fit=None, kwargs_safe=None):
        """Plot safe mask and fit mask.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axes to plot on. Default is None.
        kwargs_fit : dict, optional
            Keyword arguments passed to `~RegionNDMap.plot_mask()` for mask fit. Default is None.
        kwargs_safe : dict, optional
            Keyword arguments passed to `~RegionNDMap.plot_mask()` for mask safe. Default is None.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes object.

        Examples
        --------
        >>> # Reading a spectral dataset
        >>> from gammapy.datasets import SpectrumDatasetOnOff
        >>> filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
        >>> dataset = SpectrumDatasetOnOff.read(filename)
        >>> dataset.mask_fit = dataset.mask_safe.copy()
        >>> dataset.mask_fit.data[40:46] = False  # setting dummy mask_fit for illustration
        >>> # Plot the masks on top of the counts histogram
        >>> kwargs_safe = {"color":"green", "alpha":0.2} #optinonal arguments to configure
        >>> kwargs_fit = {"color":"pink", "alpha":0.2}
        >>> ax=dataset.plot_counts()  # doctest: +SKIP
        >>> dataset.plot_masks(ax=ax, kwargs_fit=kwargs_fit, kwargs_safe=kwargs_safe)  # doctest: +SKIP
        """
        kwargs_fit = kwargs_fit or {}
        kwargs_safe = kwargs_safe or {}

        kwargs_fit.setdefault("label", "Mask fit")
        kwargs_fit.setdefault("color", "tab:green")
        kwargs_safe.setdefault("label", "Mask safe")
        kwargs_safe.setdefault("color", "black")

        if self.mask_fit:
            self.mask_fit.plot_mask(ax=ax, **kwargs_fit)

        if self.mask_safe:
            self.mask_safe.plot_mask(ax=ax, **kwargs_safe)

        ax.legend()
        return ax

    def plot_excess(
        self, ax=None, kwargs_excess=None, kwargs_npred_signal=None, **kwargs
    ):
        """Plot excess and predicted signal.

        The error bars are computed with a symmetric assumption on the excess.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axes to plot on. Default is None.
        kwargs_excess : dict, optional
            Keyword arguments passed to `~matplotlib.axes.Axes.errorbar` for
            the excess. Default is None.
        kwargs_npred_signal : dict, optional
            Keyword arguments passed to `~matplotlib.axes.Axes.hist` for the
            predicted signal. Default is None.
        **kwargs : dict, optional
            Keyword arguments passed to both plot methods.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes object.

        Examples
        --------
        >>> #Creating a spectral dataset
        >>> from gammapy.datasets import SpectrumDatasetOnOff
        >>> from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
        >>> filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
        >>> dataset = SpectrumDatasetOnOff.read(filename)
        >>> p = PowerLawSpectralModel()
        >>> dataset.models = SkyModel(spectral_model=p)
        >>> #Plot the excess in blue and the npred in black dotted lines
        >>> kwargs_excess = {"color": "blue", "markersize":8, "marker":'s', }
        >>> kwargs_npred_signal = {"color": "black", "ls":"--"}
        >>> dataset.plot_excess(kwargs_excess=kwargs_excess, kwargs_npred_signal=kwargs_npred_signal)  # doctest: +SKIP
        """
        kwargs_excess = kwargs_excess or {}
        kwargs_npred_signal = kwargs_npred_signal or {}

        # Determine the uncertainty on the excess
        yerr = self._counts_statistic.error

        plot_kwargs = kwargs.copy()
        plot_kwargs.update(kwargs_excess)
        plot_kwargs.setdefault("label", "Excess counts")
        ax = self.excess.plot(ax, yerr=yerr, **plot_kwargs)

        plot_kwargs = kwargs.copy()
        plot_kwargs.update(kwargs_npred_signal)
        plot_kwargs.setdefault("label", "Predicted signal counts")
        self.npred_signal().plot_hist(ax, **plot_kwargs)

        ax.legend(numpoints=1)
        return ax

    def peek(self, figsize=(16, 4)):
        """Quick-look summary plots.

        This method creates a figure displaying the elements of your `SpectrumDataset`.
        For example:

        * Counts map
        * Exposure map
        * Energy dispersion matrix at the geometry center

        Parameters
        ----------
        figsize : tuple
            Size of the figure. Default is (16, 4).

        """
        _, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=figsize)

        ax1.set_title("Counts")
        self.plot_counts(ax1)
        self.plot_masks(ax=ax1)
        ax1.legend()

        ax2.set_title("Exposure")
        self.exposure.plot(ax2, ls="-", markersize=0, xerr=None)

        ax3.set_title("Energy Dispersion")

        if self.edisp is not None:
            kernel = self.edisp.get_edisp_kernel()
            kernel.plot_matrix(ax=ax3, add_cbar=True)


class SpectrumDataset(PlotMixin, MapDataset):
    """Main dataset for spectrum fitting (1D analysis).

    It bundles together binned counts, background, IRFs into `~gammapy.maps.RegionNDMap` (a Map with only one spatial bin).
    A safe mask and a fit mask can be added to exclude bins during the analysis.
    If models are assigned to it, it can compute the predicted number of counts and the statistic function,
    here the Cash statistic (see `~gammapy.stats.cash`).

    For more information see :ref:`datasets`.
    """

    tag = "SpectrumDataset"

    @classmethod
    def create(
        cls,
        geom,
        energy_axis_true=None,
        migra_axis=None,
        reference_time="2000-01-01",
        name=None,
        meta_table=None,
        **kwargs,
    ):
        """Create a `SpectrumDataset` object with zero filled maps.

        Parameters
        ----------
        geom : `~gammapy.maps.RegionGeom`
            Reference target geometry in reco energy, used for counts and background maps.
        energy_axis_true : `~gammapy.maps.MapAxis`, optional
            True energy axis used for IRF maps. Default is None.
        migra_axis : `~gammapy.maps.MapAxis`, optional
            If set, this provides the migration axis for the energy dispersion map.
            If not set, an EDispKernelMap is produced instead. Default is None.
        reference_time : `~astropy.time.Time`
            The reference time to use in GTI definition. Default is "2000-01-01".
        name : str, optional
            Name of the returned dataset. Default is None.
        meta_table : `~astropy.table.Table`, optional
            Table listing information on observations used to create the dataset.
            One line per observation for stacked datasets. Default is None.

        Returns
        -------
        empty_maps : `SpectrumDataset`
            A SpectrumDataset containing zero filled maps.

        Examples
        --------
        >>> from gammapy.datasets import SpectrumDataset
        >>> from gammapy.maps import RegionGeom, MapAxis

        >>> energy_axis = MapAxis.from_energy_bounds(1.0, 10.0, 4, unit="TeV")
        >>> energy_axis_true = MapAxis.from_energy_bounds(
        ...            0.5, 20, 10, unit="TeV", name="energy_true"
        ...        )
        >>> geom = RegionGeom.create(
        ...            region=None,
        ...            axes=[energy_axis],
        ...        )
        >>> empty = SpectrumDataset.create(geom=geom, energy_axis_true=energy_axis_true, name="empty")
        """
        if not geom.is_region:
            raise TypeError("`SpectrumDataset` is only supported for `RegionGeom`.")

        dataset = super().create(
            geom,
            energy_axis_true=energy_axis_true,
            migra_axis=migra_axis,
            reference_time=reference_time,
            name=name,
            meta_table=meta_table,
            **kwargs,
        )

        # remove PSF
        dataset.psf = None

        return dataset

    def cutout(self, *args, **kwargs):
        """Not supported for `SpectrumDataset`."""
        raise NotImplementedError("Method not supported on a spectrum dataset")

    def plot_residuals_spatial(self, *args, **kwargs):
        """Not supported for `SpectrumDataset`."""
        raise NotImplementedError("Method not supported on a spectrum dataset")

    def to_spectrum_dataset(self, *args, **kwargs):
        """Not supported for `SpectrumDataset`."""
        raise NotImplementedError("Already a Spectrum Dataset. Method not supported")


class SpectrumDatasetOnOff(PlotMixin, MapDatasetOnOff):
    """Spectrum dataset for 1D on-off likelihood fitting.

    It bundles together the binned on and off counts, the binned IRFs as well as the on and off acceptances.

    A fit mask can be added to exclude bins during the analysis.

    It uses the Wstat statistic (see `~gammapy.stats.wstat`).

    For more information see :ref:`datasets`.
    """

    tag = "SpectrumDatasetOnOff"

    def cutout(self, *args, **kwargs):
        """Not supported for `SpectrumDatasetOnOff`."""
        raise NotImplementedError("Method not supported on a spectrum dataset")

    def plot_residuals_spatial(self, *args, **kwargs):
        """Not supported for `SpectrumDatasetOnOff`."""
        raise NotImplementedError("Method not supported on a spectrum dataset")

    @classmethod
    def read(cls, filename, format=None, checksum=False, name=None, **kwargs):
        """Read from file.

        For OGIP formats, filename is the name of a PHA file. The BKG, ARF, and RMF file names must be
        set in the PHA header and the files must be present in the same folder. For details, see `OGIPDatasetReader.read`.

        For the GADF format, a MapDataset serialisation is used.

        Parameters
        ----------
        filename : `~pathlib.Path` or str
            Name of the file to read.
        format : {"ogip", "gadf"}, optional
            Format to use. If None, try to guess the format. Default is None
        checksum : bool, optional
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.
        name: str, optional
            Name of the dataset. If None, dataset name will be set to the written one. Default is None.
        kwargs : dict, optional
            Keyword arguments passed to `MapDataset.read`.
        """
        from .io import OGIPDatasetReader

        filename = make_path(filename)
        if format is None:
            with fits.open(filename) as hdulist:
                # Check for extensions in OGIP format
                if (
                    "SPECTRUM" in hdulist
                    and "OGIP" in hdulist["SPECTRUM"].header["HDUCLASS"]
                ):
                    format = "ogip"
                # Check for extensions in GADF format
                elif "COUNTS_BANDS" in hdulist:
                    format = "gadf"
                else:
                    raise ValueError(f"Cannot determine format of {filename}")

        if format == "gadf":
            return super().read(
                filename, format="gadf", checksum=checksum, name=name, **kwargs
            )
        elif format == "ogip":
            return OGIPDatasetReader(
                filename=filename, checksum=checksum, name=name
            ).read()
        else:
            raise ValueError(f"Invalid {format} serialisation format in {filename}.")

    def write(self, filename, overwrite=False, format="gadf", checksum=False):
        """Write spectrum dataset on off to file.

        Can be serialised either as a `MapDataset` with a `RegionGeom`
        following the GADF specifications, or as per the OGIP format.
        For OGIP formats specs, see `OGIPDatasetWriter`.

        Parameters
        ----------
        filename : `~pathlib.Path` or str
            Filename to write to.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        format : {"ogip", "ogip-sherpa", "gadf"}
            Format to use. Default is "gadf".
        checksum : bool
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        from .io import OGIPDatasetWriter

        if format == "gadf":
            super().write(filename=filename, overwrite=overwrite, checksum=checksum)
        elif format in ["ogip", "ogip-sherpa"]:
            creation = self.meta.creation or CreatorMetaData()
            writer = OGIPDatasetWriter(
                filename=filename,
                format=format,
                overwrite=overwrite,
                checksum=checksum,
                creation=creation,
            )
            writer.write(self)
        else:
            raise ValueError(f"{format} is not a valid serialisation format")

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create spectrum dataset from dictionary.

        Reads file from the disk as specified in the dict.

        Parameters
        ----------
        data : dict
            Dictionary containing data to create dataset from.

        Returns
        -------
        dataset : `SpectrumDatasetOnOff`
            Spectrum dataset on off.
        """
        filename = make_path(data["filename"])
        dataset = cls.read(filename=filename)
        dataset.mask_fit = None
        return dataset

    def to_dict(self):
        """Convert to dict for YAML serialization."""
        filename = f"pha_obs{self.name}.fits"
        return {"name": self.name, "type": self.tag, "filename": filename}

    @classmethod
    def from_spectrum_dataset(cls, **kwargs):
        """Create a SpectrumDatasetOnOff from a `SpectrumDataset` dataset.

        Parameters
        ----------
        dataset : `SpectrumDataset`
            Spectrum dataset defining counts, edisp, exposure etc.
        acceptance : `~numpy.array` or float
            Relative background efficiency in the on region.
        acceptance_off : `~numpy.array` or float
            Relative background efficiency in the off region.
        counts_off : `~gammapy.maps.RegionNDMap`
            Off counts spectrum. If the dataset provides a background model,
            and no off counts are defined. The off counts are deferred from
            counts_off / alpha.

        Returns
        -------
        dataset : `SpectrumDatasetOnOff`
            Spectrum dataset on off.
        """
        return cls.from_map_dataset(**kwargs)

    def to_spectrum_dataset(self, name=None):
        """Convert a SpectrumDatasetOnOff to a SpectrumDataset.

        The background model template is taken as alpha*counts_off.

        Parameters
        ----------
        name : str, optional
            Name of the new dataset. Default is None.

        Returns
        -------
        dataset : `SpectrumDataset`
            SpectrumDataset with Cash statistic.
        """
        return self.to_map_dataset(name=name).to_spectrum_dataset(on_region=None)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import time
import numpy as np
import pytest
from numpy.testing import assert_allclose
from gammapy.utils.testing import requires_dependency
from gammapy.workflow.steps import WorkflowStepBase
from gammapy.workflow.products import Products, Product


class SumStep(WorkflowStepBase):
    tag = "sum"
    parallel = True
    required_inputs = [{"name": "value"}]
    outputs_names = ["value", "log"]  # could be updated by the config in init

    def __init__(self, config=None, overwrite=False, **kwargs):
        super().__init__(config=None, overwrite=overwrite, **kwargs)

    def _sum(self):
        return np.sum(self.inputs.data)

    def _run(self):
        time.sleep(1)
        self.outputs[0].data = self._sum()
        self.outputs[1].data = "done"
        return self.outputs


def f_expected(n):
    values = []
    for _ in range(0, n - 1):
        s = np.sum([1, 2] + values)
        values.append(s)
    return s


@requires_dependency("ray")
def test_products():
    import ray

    sum1 = SumStep(name="sum1")
    assert sum1.outputs.data[0] is None
    assert sum1.outputs.data[1] is None
    assert sum1.outputs[0].step_name == "sum1"

    sum1.run(inputs=[Product(name="value", data=1), Product(name="value", data=2)])
    assert isinstance(sum1.outputs.data[0], ray.ObjectRef)
    assert isinstance(sum1.outputs.data[1], ray.ObjectRef)
    assert_allclose(sum1.inputs.data, [1, 2])

    sum1.run(inputs=sum1.outputs)
    assert_allclose(sum1.inputs.data[0], 1)
    assert_allclose(sum1.inputs.data[1], 2)
    assert isinstance(sum1.inputs.data[2], ray.ObjectRef)

    sum1.outputs.get()
    assert_allclose(sum1.outputs.data[0], 6)
    assert sum1.outputs.data[1] == "done"

    sum1.run(inputs=sum1.outputs)
    assert_allclose(sum1.inputs.data[3], 6)
    sum1.run(inputs=sum1.outputs)
    assert isinstance(sum1.inputs.data[4], ray.ObjectRef)

    sum2 = SumStep()
    assert sum2.outputs[0].step_name == sum2.name
    sum2.run(inputs=Product(name="value", data=1))
    sum2.run(inputs=Product(name="value", data=2))

    analysis_products = Products([*sum1.outputs, *sum2.outputs])
    analysis_products.get()
    results = analysis_products.select(name="value").data

    assert_allclose(results[0], f_expected(len(sum1.inputs)))
    assert_allclose(results[1], f_expected(len(sum2.inputs)))

    unique_names = analysis_products.unique_names
    assert len(unique_names) == 4
    assert unique_names[0] == f"sum1.value.{sum1.outputs[0].pid}"

    unique_names = analysis_products.select(step_name="sum1").unique_names
    for name in unique_names:
        assert "sum1" in name

    assert analysis_products.pids[0] == sum1.outputs[0].pid


def test_products_no_ray():
    sum1 = SumStep(name="sum1")
    sum1.parallel = False
    assert sum1.outputs.data[0] is None
    assert sum1.outputs.data[1] is None
    assert sum1.outputs[0].step_name == "sum1"

    sum1.run(inputs=[Product(name="value", data=1), Product(name="value", data=2)])
    assert_allclose(sum1.outputs.data[0], 3)
    assert sum1.outputs.data[1] == "done"
    assert_allclose(sum1.inputs.data, [1, 2])

    sum1.run(inputs=sum1.outputs)
    assert_allclose(sum1.inputs.data[0], 1)
    assert_allclose(sum1.inputs.data[1], 2)
    assert_allclose(sum1.inputs.data[2], 3)

    assert_allclose(sum1.outputs.data[0], 6)
    assert sum1.outputs.data[1] == "done"

    sum1.run(inputs=sum1.outputs)
    assert_allclose(sum1.inputs.data[3], 6)
    sum1.run(inputs=sum1.outputs)
    assert_allclose(sum1.inputs.data[4], 12)

    sum2 = SumStep()
    sum2.parallel = False

    assert sum2.outputs[0].step_name == sum2.name
    sum2.run(inputs=Product(name="value", data=1))
    sum2.run(inputs=Product(name="value", data=2))

    analysis_products = Products([*sum1.outputs, *sum2.outputs])
    results = analysis_products.select(name="value").data

    assert_allclose(results[0], f_expected(len(sum1.inputs)))
    assert_allclose(results[1], f_expected(len(sum2.inputs)))


def test_products_modification():
    products = Products([Product(name="value", data=1), Product(name="value", data=2)])
    products[0] = products[1]
    assert_allclose(products.data, 2)
    del products[0]
    assert len(products) == 1

    with pytest.raises(TypeError):
        products[0] = np.ones(1)
